# validated: 2016-12-10 AA e44a6e227a89 athena/java/edu/wpi/first/wpilibj/GamepadBase.java
# ----------------------------------------------------------------------------
# Copyright (c) FIRST 2016. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

from .generichid import GenericHID

__all__ = ["GamepadBase"]

class GamepadBase(GenericHID):
    """
    GamepadBase Interface.
    """

    def __init__(self, port):
        super().__init__(port)

    def getRawAxis(self, axis):
        raise NotImplementedError

    def getBumper(self, hand):
        """Is the bumper pressed.

        :param hand: which hand
        :returns: true if the bumper is pressed
        """
        raise NotImplementedError

    def getStickButton(self, hand=None):
        raise NotImplementedError

    def getRawButton(self, button):
        raise NotImplementedError

    def getPOV(self, pov=0):
        raise NotImplementedError

    def getPOVCount(self):
        raise NotImplementedError

    def getType(self):
        raise NotImplementedError

    def getName(self):
        raise NotImplementedError

    def setOutput(self, outputNumber, value):
        raise NotImplementedError

    def setOutputs(self, value):
        raise NotImplementedError

    def setRumble(self, type, value):
        raise NotImplementedError


# validated: 2016-11-26 DS e44a6e227a89 athena/java/edu/wpi/first/wpilibj/Talon.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import hal

from .livewindow import LiveWindow
from .pwmspeedcontroller import PWMSpeedController

__all__ = ["Talon"]

class Talon(PWMSpeedController):
    """
        Cross the Road Electronics (CTRE) Talon and Talon SR Speed Controller via PWM
    """

    def __init__(self, channel):
        """Constructor for a Talon (original or Talon SR)

        :param channel: The PWM channel that the Talon is attached to. 0-9 are on-board, 10-19 are on the MXP port
        :type  channel: int

        .. note ::

            The Talon uses the following bounds for PWM values. These values
            should work reasonably well for most controllers, but if users
            experience issues such as asymmetric behavior around the deadband
            or inability to saturate the controller in either direction,
            calibration is recommended.  The calibration procedure can be
            found in the Talon User Manual available from CTRE.

            - 2.037ms = full "forward"
            - 1.539ms = the "high end" of the deadband range
            - 1.513ms = center of the deadband range (off)
            - 1.487ms = the "low end" of the deadband range
            - 0.989ms = full "reverse"
        """
        super().__init__(channel)
        self.setBounds(2.037, 1.539, 1.513, 1.487, 0.989)
        self.setPeriodMultiplier(self.PeriodMultiplier.k1X)
        self.setSpeed(0.0)
        self.setZeroLatch()

        LiveWindow.addActuatorChannel("Talon", self.getChannel(), self)
        hal.report(hal.UsageReporting.kResourceType_Talon,
                   self.getChannel())



# validated: 2015-12-22 DS c3c35c6 shared/java/edu/wpi/first/wpilibj/Controller.java
__all__ = ["Controller"]

class Controller:
    """An interface for controllers. Controllers run control loops, the most
    command are PID controllers and there variants, but this includes anything
    that is controlling an actuator in a separate thread.
    """
    def enable(self):
        """Allows the control loop to run."""
        raise NotImplementedError

    def disable(self):
        """Stops the control loop from running until explicitly re-enabled by
        calling :meth:`enable`.
        """
        raise NotImplementedError



# validated: 2016-11-15 DV 7a402b4 athena/java/edu/wpi/first/wpilibj/Utility.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2016. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import hal

__all__ = ["Utility"]

class Utility:
    """Contains global utility functions"""

    @staticmethod
    def getFPGAVersion():
        """Return the FPGA Version number.

        :returns: FPGA Version number.
        :rtype: int
        """
        return hal.getFPGAVersion()

    @staticmethod
    def getFPGARevision():
        """Return the FPGA Revision number. The format of the revision is 3
        numbers.  The 12 most significant bits are the Major Revision. the
        next 8 bits are the Minor Revision. The 12 least significant bits
        are the Build Number.

        :returns: FPGA Revision number.
        :rtype: int
        """
        return hal.getFPGARevision()

    @staticmethod
    def getFPGATime():
        """Read the microsecond timer from the FPGA.

        :returns: The current time in microseconds according to the FPGA.
        :rtype: int
        """
        return hal.getFPGATime()

    @staticmethod
    def getUserButton():
        """Get the state of the "USER" button on the roboRIO.

        :returns: True if the button is currently pressed down
        :rtype: bool
        """
        return hal.getFPGAButton()



# validated: 2016-12-20 JW b51e85ae262b athena/java/edu/wpi/first/wpilibj/SPI.java

import hal
import warnings
import weakref

__all__ = ["SPI"]

def _freeSPI(port):
    hal.closeSPI(port)

class SPI:
    """Represents a SPI bus port
    
    Example usage::
    
        spi = wpilib.SPI(wpilib.SPI.Port.kOnboardCS0)
        
        # Write bytes 'text', and receive something
        data = spi.transaction(b'text')
    
    """

    class Port:
        kOnboardCS0 = 0
        kOnboardCS1 = 1
        kOnboardCS2 = 2
        kOnboardCS3 = 3
        kMXP = 4

    devices = 0
    
    @staticmethod
    def _reset():
        SPI.devices = 0

    def __init__(self, port, simPort=None):
        """Constructor

        :param port: the physical SPI port
        :type port: :class:`.SPI.Port`
        :param simPort: This must be an object that implements all of
                        the spi* functions from hal_impl that you use.
                        See ``test_spi.py`` for an example.
        """
        
        if port not in [self.Port.kOnboardCS0,
                        self.Port.kOnboardCS1,
                        self.Port.kOnboardCS2,
                        self.Port.kOnboardCS3,
                        self.Port.kMXP]:
            raise ValueError("Invalid value '%s' for SPI port" % port)
        
        if hal.HALIsSimulation():
            if simPort is None:
                # If you want more functionality, implement your own mock
                from hal_impl.spi_helpers import SPISimBase
                simPort = SPISimBase()
                
                msg = "Using stub simulator for SPI port %s" % port
                warnings.warn(msg)
            
            # Just check for basic functionality
            assert hasattr(simPort, 'initializeSPI')
            assert hasattr(simPort, 'closeSPI')
            
            self._port = (simPort, port)
        else:
            self._port = port
        
        self.bitOrder = False
        self.clockPolarity = False
        self.dataOnTrailing = False

        hal.initializeSPI(self._port)
        self.__finalizer = weakref.finalize(self, _freeSPI, self._port)

        SPI.devices += 1
        hal.report(hal.UsageReporting.kResourceType_SPI, SPI.devices)
    
    @property
    def port(self):
        if not self.__finalizer.alive:
            raise ValueError("Cannot use SPI after free() has been called")
        return self._port

    def free(self):
        self.__finalizer()

    def setClockRate(self, hz):
        """Configure the rate of the generated clock signal.
        The default value is 500,000 Hz.
        The maximum value is 4,000,000 Hz.

        :param hz: The clock rate in Hertz.
        """
        hal.setSPISpeed(self.port, hz)

    def setMSBFirst(self):
        """Configure the order that bits are sent and received on the wire
        to be most significant bit first.
        """
        self.bitOrder = True
        hal.setSPIOpts(self.port, self.bitOrder, self.dataOnTrailing,
                       self.clockPolarity)

    def setLSBFirst(self):
        """Configure the order that bits are sent and received on the wire
        to be least significant bit first.
        """
        self.bitOrder = False
        hal.setSPIOpts(self.port, self.bitOrder, self.dataOnTrailing,
                       self.clockPolarity)

    def setClockActiveLow(self):
        """Configure the clock output line to be active low.
        This is sometimes called clock polarity high or clock idle high.
        """
        self.clockPolarity = True
        hal.setSPIOpts(self.port, self.bitOrder, self.dataOnTrailing,
                       self.clockPolarity)

    def setClockActiveHigh(self):
        """Configure the clock output line to be active high.
        This is sometimes called clock polarity low or clock idle low.
        """
        self.clockPolarity = False
        hal.setSPIOpts(self.port, self.bitOrder, self.dataOnTrailing,
                       self.clockPolarity)

    def setSampleDataOnFalling(self):
        """Configure that the data is stable on the falling edge and the data
        changes on the rising edge.
        """
        self.dataOnTrailing = True
        hal.setSPIOpts(self.port, self.bitOrder, self.dataOnTrailing,
                       self.clockPolarity)

    def setSampleDataOnRising(self):
        """Configure that the data is stable on the rising edge and the data
        changes on the falling edge.
        """
        self.dataOnTrailing = False
        hal.setSPIOpts(self.port, self.bitOrder, self.dataOnTrailing,
                       self.clockPolarity)

    def setChipSelectActiveHigh(self):
        """Configure the chip select line to be active high.
        """
        hal.setSPIChipSelectActiveHigh(self.port)

    def setChipSelectActiveLow(self):
        """Configure the chip select line to be active low.
        """
        hal.setSPIChipSelectActiveLow(self.port)

    def write(self, dataToSend):
        """Write data to the slave device.  Blocks until there is space in the
        output FIFO.

        If not running in output only mode, also saves the data received
        on the MISO input during the transfer into the receive FIFO.

        :param dataToSend: Data to send
        :type dataToSend: iterable of bytes
        
        :returns: Number of bytes written
        
        Usage::
        
            # send byte string
            writeCount = spi.write(b'stuff')
            
            # send list of integers
            writeCount = spi.write([0x01, 0x02])
        """
        return hal.writeSPI(self.port, dataToSend)

    def read(self, initiate, size):
        """Read a word from the receive FIFO.

        Waits for the current transfer to complete if the receive FIFO is
        empty.

        If the receive FIFO is empty, there is no active transfer, and
        initiate is False, errors.

        :param initiate: If True, this function pushes "0" into the
            transmit buffer and initiates a transfer.  If False, this function
            assumes that data is already in the receive FIFO from a previous
            write.
        :param size: Number of bytes to read.

        :returns: received data bytes
        """
        if initiate:
            return hal.transactionSPI(self.port, [0]*size)
        else:
            return hal.readSPI(self.port, size)

    def transaction(self, dataToSend):
        """Perform a simultaneous read/write transaction with the device

        :param dataToSend: The data to be written out to the device
        :type dataToSend: iterable of bytes

        :returns: data received from the device
        
        Usage::
        
            # send byte string
            data = spi.transaction(b'stuff')
            
            # send list of integers
            data = spi.transaction([0x01, 0x02])
        """
        return hal.transactionSPI(self.port, dataToSend)
    
    
    def initAccumulator(self, period, cmd, xfer_size,
                        valid_mask, valid_value,
                        data_shift, data_size,
                        is_signed, big_endian):
        """Initialize the accumulator.
        
        :param period: Time between reads
        :param cmd: SPI command to send to request data
        :param xfer_size: SPI transfer size, in bytes
        :param valid_mask: Mask to apply to received data for validity checking
        :param valid_data: After valid_mask is applied, required matching value for
                           validity checking
        :param data_shift: Bit shift to apply to received data to get actual data
                           value
        :param data_size: Size (in bits) of data field
        :param is_signed: Is data field signed?
        :param big_endian: Is device big endian?
        """
        hal.initSPIAccumulator(self.port, int(period*1.0e6), cmd,
                               xfer_size, valid_mask, valid_value, data_shift,
                               data_size, is_signed, big_endian)

    def freeAccumulator(self):
        """Frees the accumulator."""
        hal.freeSPIAccumulator(self.port)

    def resetAccumulator(self):
        """Resets the accumulator to zero."""
        hal.resetSPIAccumulator(self.port)

    def setAccumulatorCenter(self, center):
        """Set the center value of the accumulator.
        
        The center value is subtracted from each value before it is added to the accumulator. This
        is used for the center value of devices like gyros and accelerometers to make integration work
        and to take the device offset into account when integrating.
        """
        hal.setSPIAccumulatorCenter(self.port, center)

    def setAccumulatorDeadband(self, deadband):
        """Set the accumulator's deadband."""
        hal.setSPIAccumulatorDeadband(self.port, deadband)

    def getAccumulatorLastValue(self):
        """Read the last value read by the accumulator engine."""

        return hal.getSPIAccumulatorLastValue(self.port);

    def getAccumulatorValue(self):
        """Read the accumulated value.
        
        :returns: The 64-bit value accumulated since the last Reset().
        """
        return hal.getSPIAccumulatorValue(self.port)
    
    def getAccumulatorCount(self):
        """Read the number of accumulated values.
        
        Read the count of the accumulated values since the accumulator was last Reset().
        
        :returns: The number of times samples from the channel were accumulated.
        """
        return hal.getSPIAccumulatorCount(self.port)
    
    def getAccumulatorAverage(self):
        """Read the average of the accumulated value.
        
        :returns: The accumulated average value (value / count).
        """
        return hal.getSPIAccumulatorAverage(self.port)
    
    def getAccumulatorOutput(self):
        """Read the accumulated value and the number of accumulated values atomically.
        
        This function reads the value and count atomically.
        This can be used for averaging.
        
        :returns: tuple of (value, count)
        """
        return hal.getSPIAccumulatorOutput(self.port)    
 



# validated: 2016-11-17 AA 64ebe7f shared/java/edu/wpi/first/wpilibj/smartdashboard/SmartDashboard.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import hal

__all__ = ["SmartDashboard"]

class SmartDashboard:
    """The bridge between robot programs and the SmartDashboard on the laptop

    When a value is put into the SmartDashboard, it pops up on the
    SmartDashboard on the remote host. Users can put values into and get values
    from the SmartDashboard.
    
    These values can also be accessed by a NetworkTables client via the
    'SmartDashboard' table::
    
        from networktables import NetworkTable
        sd = NetworkTable.getTable('SmartDashboard')
        
        # sd.putXXX and sd.getXXX work as expected here
    
    """
    # The NetworkTable used by SmartDashboard
    table = None
    # A table linking tables in the SmartDashboard to the SmartDashboardData
    # objects they came from.
    tablesToData = {}

    class _defaultValueSentry:
        pass
    
    @classmethod
    def _reset(cls):
        cls.tablesToData = {}
        cls.table = None

    @classmethod
    def getTable(cls):
        if cls.table is None:
            from networktables import NetworkTable
            cls.table = NetworkTable.getTable("SmartDashboard")
            hal.report(hal.UsageReporting.kResourceType_SmartDashboard,
                       hal.UsageReporting.kSmartDashboard_Instance)
        return cls.table

    @classmethod
    def putData(cls, *args, **kwargs):
        """Maps the specified key to the specified value in this table.
        The value can be retrieved by calling the get method with a key that
        is equal to the original key.

        Two argument formats are supported: key, data:
        
        :param key: the key (cannot be None)
        :type  key: str
        :param data: the value

        Or the single argument "value":
        
        :param value: the named value (getName is called to retrieve the value)
        """
        # NOTE: mix of args and kwargs not allowed
        if kwargs and not args:
            if "value" in kwargs:
                data = kwargs["value"]
                key = data.getName()
            else:
                key = kwargs["key"]
                data = kwargs["data"]
        elif len(args) == 1 and not kwargs:
            data = args[0]
            key = data.getName()
        elif len(args) == 2 and not kwargs:
            key, data = args
        else:
            raise ValueError("only key, data or value accepted")

        table = cls.getTable()
        dataTable = table.getSubTable(key)
        dataTable.putString("~TYPE~", data.getSmartDashboardType())
        data.initTable(dataTable)
        cls.tablesToData[dataTable] = data

    @classmethod
    def getData(cls, key):
        """Returns the value at the specified key.
        
        :param key: the key
        :type  key: str
        :returns: the value
        
        :raises: :exc:`KeyError` if the key doesn't exist
        """
        table = cls.getTable()
        subtable = table.getSubTable(key)
        data = cls.tablesToData.get(subtable)
        if data is None:
            raise KeyError("SmartDashboard data does not exist: '%s'" % key)
        return data

    @classmethod
    def containsKey(cls, key):
        """Checks the table and tells if it contains the specified key.

        :param key: key the key to search for
        :return: true if the table as a value assigned to the given key
        """
        table = cls.getTable()
        return table.containsKey(key)

    @classmethod
    def getKeys(cls, types=0):
        """Get array of keys in the table.

        :param types: bitmask of types; 0 is treated as a "don't care".
        :return: keys currently in the table
        """
        table = cls.getTable()
        return table.getKeys(types)

    @classmethod
    def setPersistent(cls, key):
        """Makes a key's value persistent through program restarts.
        The key cannot be null.

        :param key: the key name
        """
        table = cls.getTable()
        table.setPersistent(key)

    @classmethod
    def clearPersistent(cls, key):
        """Stop making a key's value persistent through program restarts.
        The key cannot be null.

        :param key: the key name
        """
        table = cls.getTable()
        table.clearPersistent(key)

    @classmethod
    def isPersistent(cls, key):
        """Returns whether the value is persistent through program restarts.
        The key cannot be null.

        :param key: the key name
        :return: True if the value is persistent.
        """
        table = cls.getTable()
        return table.isPersistent(key)

    @classmethod
    def setFlags(cls, key, flags):
        """Sets flags on the specified key in this table. The key can
        not be null.

        :param key: the key name
        :param flags: the flags to set (bitmask)
        """
        table = cls.getTable()
        table.setFlags(key, flags)

    @classmethod
    def clearFlags(cls, key, flags):
        """Clears flags on the specified key in this table. The key can
        not be null.

        :param key: the key name
        :param flags: the flags to clear (bitmask)
        """
        table = cls.getTable()
        table.clearFlags()

    @classmethod
    def getFlags(cls, key):
        """ Returns the flags for the specified key.

        :param key: the key name
        :return: the flags, or 0 if the key is not defined
        """
        table = cls.getTable()
        return table.getFlags(key)

    @classmethod
    def delete(cls, key):
        """Deletes the specified key in this table. The key can
        not be null.

        :param key: the key name
        """
        table = cls.getTable()
        table.delete(key)

    @classmethod
    def putBoolean(cls, key, value):
        """Put a boolean in the table.

        :param key: the key to be assigned to
        :param value: the value that will be assigned
        :return False if the table key already exists with a different type
        """
        table = cls.getTable()
        return table.putBoolean(key, value)

    @classmethod
    def setDefaultBoolean(cls, key, defaultValue):
        """Gets the current value in the table, setting it if it does not exist.
        
        :param key: the key
        :param defaultValue: the default value to set if key doens't exist.
        :returns: False if the table key exists with a different type
        """
        table = cls.getTable()
        return table.setDefaultBoolean(key, defaultValue)

    @classmethod
    def getBoolean(cls, key, defaultValue=_defaultValueSentry):
        """Returns the boolean the key maps to. If the key does not exist or is of
        different type, it will return the default value; if that is not provided,
        it will throw a :exc:`KeyError`.
        
        Calling this method without passing defaultValue is deprecated.

        :param key: the key to look up
        :type  key: str
        :param defaultValue: returned if the key doesn't exist
        :returns: the value associated with the given key or the given default value
                  if there is no value associated with the key
        
        :raises: :exc:`KeyError` if the key doesn't exist and defaultValue
                 is not provided.
        """
        table = cls.getTable()
        if defaultValue is cls._defaultValueSentry:
            return table.getBoolean(key)
        else:
            return table.getBoolean(key, defaultValue)

    @classmethod
    def putNumber(cls, key, value):
        """Put a number in the table.
        
        :param key: the key to be assigned to
        :param value: the value that will be assigned
        :returns: False if the table key already exists with a different type
        """
        table = cls.getTable()
        return table.putNumber(key, value)

    @classmethod
    def setDefaultNumber(cls, key, defaultValue):
        """Gets the current value in the table, setting it if it does not exist.
        
        :param key: the key
        :param defaultValue: the default value to set if key doens't exist.
        :returns: False if the table key exists with a different type
        """
        table = cls.getTable()
        return table.setDefaultNumber(key, defaultValue)

    @classmethod
    def getNumber(cls, key, defaultValue=_defaultValueSentry):
        """Returns the number the key maps to. If the key does not exist or is of
        different type, it will return the default value; if that is not provided,
        it will throw a :exc:`KeyError`.
        
        Calling this method without passing defaultValue is deprecated.

        :param key: the key to look up
        :type  key: str
        :param defaultValue: returned if the key doesn't exist
        :returns: the value associated with the given key or the given default value
                  if there is no value associated with the key

        :raises: :exc:`KeyError` if the key doesn't exist and defaultValue
                 is not provided.
        """
        table = cls.getTable()
        if defaultValue is cls._defaultValueSentry:
            return table.getNumber(key)
        else:
            return table.getNumber(key, defaultValue)

    @classmethod
    def putString(cls, key, value):
        """Put a string in the table.
        
        :param key: the key to be assigned to
        :param value: the value that will be assigned
        :returns: False if the table key already exists with a different type
        """
        table = cls.getTable()
        return table.putString(key, value)

    @classmethod
    def setDefaultString(cls, key, defaultValue):
        """Gets the current value in the table, setting it if it does not exist.
        
        :param key: the key
        :param defaultValue: the default value to set if key doens't exist.
        :returns: False if the table key exists with a different type
        """
        table = cls.getTable()
        return table.setDefaultString(key, defaultValue)


    @classmethod
    def getString(cls, key, defaultValue=_defaultValueSentry):
        """Returns the string the key maps to. If the key does not exist or is of
        different type, it will return the default value; if that is not provided,
        it will throw a :exc:`KeyError`.
        
        Calling this method without passing defaultValue is deprecated.

        :param key: the key to look up
        :type  key: str
        :param defaultValue: returned if the key doesn't exist
        :returns: the value associated with the given key or the given default value
                  if there is no value associated with the key

        :raises: :exc:`KeyError` if the key doesn't exist and defaultValue
                 is not provided.
        """
        table = cls.getTable()
        if defaultValue is cls._defaultValueSentry:
            return table.getString(key)
        else:
            return table.getString(key, defaultValue)

    @classmethod
    def putBooleanArray(cls, key, value):
        """Put a boolean array in the table.
        
        :param key: the key to be assigned to
        :param value: the value that will be assigned
        :returns: False if the table key already exists with a different type
        """
        table = cls.getTable()
        return table.putBooleanArray(key, value)

    @classmethod
    def setDefaultBooleanArray(cls, key, defaultValue):
        """Gets the current value in the table, setting it if it does not exist.
        
        :param key: the key
        :param defaultValue: the default value to set if key doens't exist.
        :returns: False if the table key exists with a different type
        """
        table = cls.getTable()
        return table.setDefaultBooleanArray(key, defaultValue)


    @classmethod
    def getBooleanArray(cls, key, defaultValue=_defaultValueSentry):
        """Returns the boolean array the key maps to. If the key does not exist or is of
        different type, it will return the default value; if that is not provided,
        it will throw a :exc:`KeyError`.
        
        Calling this method without passing defaultValue is deprecated.

        :param key: the key to look up
        :type  key: str
        :param defaultValue: returned if the key doesn't exist
        :returns: the value associated with the given key or the given default value
                  if there is no value associated with the key

        :raises: :exc:`KeyError` if the key doesn't exist and defaultValue
                 is not provided.
        """
        table = cls.getTable()
        if defaultValue is cls._defaultValueSentry:
            return table.getBooleanArray(key)
        else:
            return table.getBooleanArray(key, defaultValue)

    @classmethod
    def putNumberArray(cls, key, value):
        """Put a number array in the table.
        
        :param key: the key to be assigned to
        :param value: the value that will be assigned
        :returns: False if the table key already exists with a different type
        """
        table = cls.getTable()
        return table.putNumberArray(key, value)

    @classmethod
    def setDefaultNumberArray(cls, key, defaultValue):
        """Gets the current value in the table, setting it if it does not exist.
        
        :param key: the key
        :param defaultValue: the default value to set if key doens't exist.
        :returns: False if the table key exists with a different type
        """
        table = cls.getTable()
        return table.setDefaultNumberArray(key, defaultValue)


    @classmethod
    def getNumberArray(cls, key, defaultValue=_defaultValueSentry):
        """Returns the number array the key maps to. If the key does not exist or is of
        different type, it will return the default value; if that is not provided,
        it will throw a :exc:`KeyError`.
        
        Calling this method without passing defaultValue is deprecated.

        :param key: the key to look up
        :type  key: str
        :param defaultValue: returned if the key doesn't exist
        :returns: the value associated with the given key or the given default value
                  if there is no value associated with the key

        :raises: :exc:`KeyError` if the key doesn't exist and defaultValue
                 is not provided.
        """
        table = cls.getTable()
        if defaultValue is cls._defaultValueSentry:
            return table.getNumberArray(key)
        else:
            return table.getNumberArray(key, defaultValue)

    @classmethod
    def putRaw(cls, key, value):
        """Put a raw value (byte array) in the table.
        
        :param key: the key to be assigned to
        :param value: the value that will be assigned
        :returns: False if the table key already exists with a different type
        """
        table = cls.getTable()
        return table.putRaw(key, value)

    @classmethod
    def setDefaultRaw(cls, key, defaultValue):
        """Gets the current value in the table, setting it if it does not exist.
        
        :param key: the key
        :param defaultValue: the default value to set if key doens't exist.
        :returns: False if the table key exists with a different type
        """
        table = cls.getTable()
        return table.setDefaultNumberArray(key, defaultValue)


    @classmethod
    def getRaw(cls, key, defaultValue=_defaultValueSentry):
        """Returns the raw value (byte array) the key maps to. If the key does not exist or is of
        different type, it will return the default value; if that is not provided,
        it will throw a :exc:`KeyError`.
        
        Calling this method without passing defaultValue is deprecated.

        :param key: the key to look up
        :type  key: str
        :param defaultValue: returned if the key doesn't exist
        :returns: the value associated with the given key or the given default value
                  if there is no value associated with the key

        :raises: :exc:`KeyError` if the key doesn't exist and defaultValue
                 is not provided.
        """
        table = cls.getTable()
        if defaultValue is cls._defaultValueSentry:
            return table.getNumberArray(key)
        else:
            return table.getNumberArray(key, defaultValue)



    # Deprecated Methods
    putInt = putNumber
    getInt = getNumber
    putDouble = putNumber
    getDouble = getNumber



# validated: 2015-12-24 DS 6d854af athena/java/edu/wpi/first/wpilibj/ControllerPower.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import hal

__all__ = ['ControllerPower']

class ControllerPower:
    """Provides access to power levels on the roboRIO"""

    @staticmethod
    def getInputVoltage():
        """
            Get the input voltage to the robot controller
            
            :returns: The controller input voltage value in Volts
            :rtype: float
        """
        return hal.getVinVoltage()
    
    @staticmethod
    def getInputCurrent():
        """
            Get the input current to the robot controller
            
            :returns: The controller input current value in Amps
            :rtype: float
        """
        return hal.getVinCurrent()
    
    @staticmethod
    def getVoltage3V3():
        """
            Get the voltage of the 3.3V rail
            
            :returns: The controller 3.3V rail voltage value in Volts
            :rtype: float
        """
        return hal.getUserVoltage3V3()
    
    @staticmethod
    def getCurrent3V3():
        """
            Get the current output of the 3.3V rail
            
            :returns: The controller 3.3V rail output current value in Amps
            :rtype: float
        """
        return hal.getUserCurrent3V3()
    
    @staticmethod
    def getEnabled3V3():
        """
            Get the enabled state of the 3.3V rail. The rail may be
            disabled due to a controller brownout, a short circuit on the
            rail, or controller over-voltage
            
            :returns: True if enabled, False otherwise
            :rtype: bool
        """
        return hal.getUserActive3V3();
    
    @staticmethod
    def getFaultCount3V3():
        """
            Get the count of the total current faults on the 3.3V rail since
            the controller has booted
            
            :returns: The number of faults
            :rtype: int
        """
        return hal.getUserCurrentFaults3V3()
    
    @staticmethod
    def getVoltage5V():
        """
            Get the voltage of the 5V rail
            
            :returns: The controller 5V rail voltage value in Volts
            :rtype: float
        """
        return hal.getUserVoltage5V()
    
    @staticmethod
    def getCurrent5V():
        """
            Get the current output of the 5V rail
            
            :returns: The controller 5V rail output current value in Amps
            :rtype: float
        """
        return hal.getUserCurrent5V()
    
    @staticmethod
    def getEnabled5V():
        """
            Get the enabled state of the 5V rail. The rail may be disabled
            due to a controller brownout, a short circuit on the rail, or
            controller over-voltage
            
            :returns: True if enabled, False otherwise
            :rtype: bool
        """
        return hal.getUserActive5V()
    
    @staticmethod
    def getFaultCount5V():
        """
            Get the count of the total current faults on the 5V rail since
            the controller has booted
            
            :returns: The number of faults
            :rtype: int
        """
        return hal.getUserCurrentFaults5V()
    
    @staticmethod
    def getVoltage6V():
        """
            Get the voltage of the 6V rail
            
            :returns: The controller 6V rail voltage value in Volts
            :rtype: float
        """
        return hal.getUserVoltage6V()
    
    @staticmethod
    def getCurrent6V():
        """
            Get the current output of the 6V rail
            
            :returns: The controller 6V rail output current value in Amps
            :rtype: float
        """
        return hal.getUserCurrent6V()
    
    @staticmethod
    def getEnabled6V():
        """
            Get the enabled state of the 6V rail. The rail may be disabled
            due to a controller brownout, a short circuit on the rail, or
            controller over-voltage
            
            :returns: True if enabled, False otherwise
            :rtype: bool
        """
        return hal.getUserActive6V()
    
    @staticmethod
    def getFaultCount6V():
        """
            Get the count of the total current faults on the 6V rail since
            the controller has booted
            
            :returns: The number of faults
            :rtype: int
        """
        return hal.getUserCurrentFaults6V()




"""
    Classes in this package are used to interface various types of buttons
    to a command-based robot.
    
    If you are not using the Command framework, you can ignore these classes.
"""

from .button import *
from .internalbutton import *
from .joystickbutton import *
from .networkbutton import *
from .trigger import *



# validated: 2016-12-21 JW 8cec94869980 athena/java/edu/wpi/first/wpilibj/AnalogTriggerOutput.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import hal

from .digitalsource import DigitalSource

__all__ = ["AnalogTriggerOutput"]

class AnalogTriggerOutput(DigitalSource):
    """Represents a specific output from an :class:`.AnalogTrigger`
    
    This class is used to get the current output value and also as a
    :class:`.DigitalSource` to provide routing of an output to digital
    subsystems on the FPGA such as :class:`.Counter`, :class:`.Encoder:,
    and :class:`.Interrupt`.

    The TriggerState output indicates the primary output value of the trigger.
    If the analog signal is less than the lower limit, the output is False. If
    the analog value is greater than the upper limit, then the output is True.
    If the analog value is in between, then the trigger output state maintains
    its most recent value.

    The InWindow output indicates whether or not the analog signal is inside
    the range defined by the limits.

    The RisingPulse and FallingPulse outputs detect an instantaneous transition
    from above the upper limit to below the lower limit, and vise versa. These
    pulses represent a rollover condition of a sensor and can be routed to an
    up / down couter or to interrupts. Because the outputs generate a pulse,
    they cannot be read directly. To help ensure that a rollover condition is
    not missed, there is an average rejection filter available that operates on
    the upper 8 bits of a 12 bit number and selects the nearest outlyer of 3
    samples.  This will reject a sample that is (due to averaging or sampling)
    errantly between the two limits. This filter will fail if more than one
    sample in a row is errantly in between the two limits. You may see this
    problem if attempting to use this feature with a mechanical rollover
    sensor, such as a 360 degree no-stop potentiometer without signal
    conditioning, because the rollover transition is not sharp / clean enough.
    Using the averaging engine may help with this, but rotational speeds of the
    sensor will then be limited.
    """

    def __init__(self, trigger, outputType):
        """Create an object that represents one of the four outputs from an
        analog trigger.

        Because this class derives from DigitalSource, it can be passed into
        routing functions for Counter, Encoder, etc.

        :param trigger: The trigger for which this is an output.
        :param outputType: An enum that specifies the output on the trigger
            to represent.
        """
        self.trigger = trigger
        self.outputType = outputType

        hal.report(hal.UsageReporting.kResourceType_AnalogTriggerOutput,
                      trigger.index, outputType)

    def free(self):
        if self.interrupt is not None:
            self.cancelInterrupts()

    def get(self):
        """Get the state of the analog trigger output.

        :returns: The state of the analog trigger output.
        :rtype: :class:`.AnalogTriggerType`
        """
        return hal.getAnalogTriggerOutput(self.trigger.port, self.outputType)

    def getPortHandleForRouting(self):
        return self.trigger.port

    def getChannel(self):
        return self.trigger.index

    def getAnalogTriggerTypeForRouting(self):
        return self.outputType

    class AnalogTriggerType:
        """Defines the state in which the :class:`.AnalogTrigger` triggers"""
        kInWindow = hal.AnalogTriggerType.kInWindow
        kState = hal.AnalogTriggerType.kState
        kRisingPulse = hal.AnalogTriggerType.kRisingPulse
        kFallingPulse = hal.AnalogTriggerType.kFallingPulse



# novalidate

import logging
import pprint

# TODO: Make these configurable
log_datefmt = "%H:%M:%S"
log_format = "%(asctime)s:%(msecs)03d %(levelname)-8s: %(name)-20s: %(message)s"

def configure_logging(verbose):
    
    formatter = VerboseExceptionFormatter(fmt=log_format,
                                          datefmt=log_datefmt)
    
    # console logging
    handler = logging.StreamHandler()
    handler.setFormatter(formatter)
    
    logging.root.addHandler(handler)
    logging.root.setLevel(logging.DEBUG if verbose else logging.INFO)



MAX_VARS_LINES = 30
MAX_LINE_LENGTH = 100


class VerboseExceptionFormatter(logging.Formatter):
    '''
    Taken from http://word.bitly.com/post/69080588278/logging-locals
    '''
 
    def __init__(self, log_locals_on_exception=True, *args, **kwargs):
        self._log_locals = log_locals_on_exception
        super(VerboseExceptionFormatter, self).__init__(*args, **kwargs)
 
    def formatException(self, exc_info):
        # First get the original formatted exception.
        exc_text = super(VerboseExceptionFormatter, self).formatException(exc_info)
        if not self._log_locals:
            return exc_text
        # Now we're going to format and add the locals information.
        output_lines = [exc_text, '\n']
        tb = exc_info[2]  # This is the outermost frame of the traceback.
        while tb.tb_next:
            tb = tb.tb_next  # Zoom to the innermost frame.
        output_lines.append('Locals at innermost frame:\n')
        locals_text = pprint.pformat(tb.tb_frame.f_locals, indent=2)
        locals_lines = locals_text.split('\n')
        if len(locals_lines) > MAX_VARS_LINES:
            locals_lines = locals_lines[:MAX_VARS_LINES]
            locals_lines[-1] = '...'
        output_lines.extend(
            line[:MAX_LINE_LENGTH - 3] + '...' if len(line) > MAX_LINE_LENGTH else line
            for line in locals_lines)
        output_lines.append('\n')
        return '\n'.join(output_lines)
    




# validated: 2017-01-12 DS 8f67f2c24cb9 athena/java/edu/wpi/first/wpilibj/XboxController.java
import hal
from .driverstation import DriverStation
from .interfaces.gamepadbase import GamepadBase
from wpilib.interfaces.generichid import GenericHID


class XboxController(GamepadBase):
    """
        Handle input from Xbox 360 or Xbox One controllers connected to the Driver Station.

        This class handles Xbox input that comes from the Driver Station. Each time a value is
        requested the most recent value is returned. There is a single class instance for each controller
        and the mapping of ports to hardware buttons depends on the code in the Driver Station.
     """

    def __init__(self, port):
        """Construct an instance of an XBoxController. The XBoxController index is the USB port on the Driver Station.

        :param port: The port on the Driver Station that the joystick is plugged into
        """

        super().__init__(port)
        self.ds = DriverStation.getInstance()

        self.outputs = 0
        self.leftRumble = 0
        self.rightRumble = 0

        hal.report(hal.UsageReporting.kResourceType_Joystick, port)

    def getX(self, hand):
        """Get the X axis value of the controller.

        :param hand: Side of controller whose value should be returned
        :return: The X axis value of the controller
        :rtype: float
        """
        if hand == GenericHID.Hand.kLeft:
            return self.getRawAxis(0)
        else:
            return self.getRawAxis(4)

    def getY(self, hand):
        """Get the Y axis value of the controller.

        :param hand: Side of controller whose value should be returned
        :return: The Y axis value of the controller
        :rtype: float
        """
        if hand == GenericHID.Hand.kLeft:
            return self.getRawAxis(1)
        else:
            return self.getRawAxis(5)

    def getRawAxis(self, axis):
        """Get the value of the axis

        :param axis: The axis to read, starting at 0
        :return: The value of the axis
        :rtype: float
        """
        return self.ds.getStickAxis(self.port, axis)

    def getBumper(self, hand):
        """Read the values of the bumper button on the controller.

        :param hand: Side of controller whose value should be returned.
        :return: The state of the button
        :rtype: boolean
        """
        if hand == GenericHID.Hand.kLeft:
            return self.getRawButton(5)
        else:
            return self.getRawButton(6)

    def getRawButton(self, button):
        """Get the buttom value (starting at button 1)

        :param button: The button number to be read (starting at 1)
        :return: The state of the button
        :rtype: boolean
        """
        return self.ds.getStickButton(self.port, button)

    def getTriggerAxis(self, hand):
        """Get the trigger axis value of the controller.

        :param hand: Side of controller whose value should be returned
        :return: The trigger axis value of the controller
        :rtype: float
        """
        if hand == GenericHID.Hand.kLeft:
            return self.getRawAxis(2)
        else:
            return self.getRawAxis(3)

    def getAButton(self):
        """Read the value of the A button on the controller

        :return: The state of the A button
        :rtype: boolean
        """
        return self.getRawButton(1)

    def getBButton(self):
        """Read the value of the B button on the controller

        :return: The state of the B button
        :rtype: boolean
        """
        return self.getRawButton(2)

    def getXButton(self):
        """Read the value of the X button on the controller

        :return: The state of the X button
        :rtype: boolean
        """
        return self.getRawButton(3)

    def getYButton(self):
        """Read the value of the Y button on the controller

        :return: The state of the Y button
        :rtype: boolean
        """
        return self.getRawButton(4)

    def getStickButton(self, hand):
        """Read the values of the stick button on the controller

        :param hand: Side of the controller whose value should be returned
        :return: The state of the button
        :rtype: boolean
        """
        if hand == GenericHID.Hand.kLeft:
            return self.getRawButton(9)
        else:
            return self.getRawButton(10)

    def getBackButton(self):
        """Read the value of the back button on the controller

        :return: The state of the back button
        :rtype: boolean
        """
        return self.getRawButton(7)

    def getStartButton(self):
        """Read the value of the start button on the controller

        :return: The state of the start button
        :rtype: boolean
        """
        return self.getRawButton(8)

    def getPOV(self, pov):
        return self.ds.getStickPOV(self.port, pov)

    def getPOVCount(self):
        return self.ds.getPOVCount(self.port)

    def getType(self):
        return self.ds.getJoystickType(self.port)

    def getName(self):
        return self.ds.getJoystickName(self.port)

    def setOutput(self, outputNumber, value):
        self.outputs = ((self.outputs & ~(1 << (outputNumber - 1)))
                        | ((1 if value else 0) << (outputNumber - 1)))
        hal.setJoystickOutputs(self.port, self.outputs, self.leftRumble, self.rightRumble)

    def setOutputs(self, value):
        self.outputs = value
        hal.setJoystickOutputs(self.port, self.outputs, self.leftRumble, self.rightRumble)

    def setRumble(self, type_, value):
        if value < 0:
            value = 0
        if value > 1:
            value = 1

        if type_ == GenericHID.RumbleType.kLeftRumble:
            self.leftRumble = int(value * 65535)
        else:
            self.rightRumble = int(value * 65535)

        hal.setJoystickOutputs(
            self.port,
            self.outputs,
            self.leftRumble,
            self.rightRumble)



# validated: 2017-01-13 DS 7a049c29bdb7 shared/java/edu/wpi/first/wpilibj/command/Command.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2016. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

from .scheduler import Scheduler
from ..robotstate import RobotState
from ..sendable import Sendable
from ..timer import Timer

import threading

__all__ = ["Command"]

class Command(Sendable):
    """The Command class is at the very core of the entire command framework.
    Every command can be started with a call to start().
    Once a command is started it will call :meth:`initialize`, and then
    will repeatedly call :meth:`execute` until :meth:`isFinished` returns True.
    Once it does, :meth:`end` will be called.
    
    However, if at any point while it is running :meth:`cancel` is called, then
    the command will be stopped and :meth:`interrupted` will be called.
    
    If a command uses a :class:`.Subsystem`, then it should specify that it
    does so by calling the :meth:`requires` method in its constructor.
    Note that a Command may have multiple requirements, and :meth:`requires`
    should be called for each one.
    
    If a command is running and a new command with shared requirements is
    started, then one of two things will happen.  If the active command is
    interruptible, then :meth:`cancel` will be called and the command will be removed
    to make way for the new one.  If the active command is not interruptible,
    the other one will not even be started, and the active one will continue
    functioning.
    
    .. seealso:: :class:`.Subsystem`, :class:`.CommandGroup`
    """

    def __init__(self, name=None, timeout=None):
        """Creates a new command.
        
        :param name: The name for this command; if unspecified or None,
                     The name of this command will be set to its class name.
        :param timeout: The time (in seconds) before this command "times out".
                        Default is no timeout.  See isTimedOut().
        """
        self.mutex = threading.RLock()

        # The name of this command
        if name is None:
            self.name = self.__class__.__name__
        else:
            self.name = name

        # The time (in seconds) before this command "times out" (or None if no
        # timeout)
        if timeout is not None and timeout < 0:
            raise ValueError("Timeout must not be negative")
        self.timeout = timeout

        # The time since this command was initialized
        self.startTime = None
        # Whether or not this command has been initialized
        self.initialized = False
        # The requirements (or null if no requirements)
        self.requirements = set()
        # Whether or not it is running
        self.running = False
        # Whether or not it is interruptible
        self.interruptible = True
        # Whether or not it has been canceled
        self.canceled = False
        # Whether or not it has been locked
        self.locked = False
        # Whether this command should run when the robot is disabled
        self.runWhenDisabled = False
        # The CommandGroup this is in
        self.parent = None

    def getName(self):
        """Returns the name of this command. If no name was specified
        in the constructor, then the default is the name of the class.
        
        :returns: the name of this command
        """
        return self.name

    def setTimeout(self, seconds):
        """Sets the timeout of this command.
        
        :param seconds: the timeout (in seconds)
        
        :see: :meth:`isTimedOut`
        """
        if seconds < 0:
            raise ValueError("Seconds must be positive.")
        with self.mutex:
            self.timeout = seconds

    def timeSinceInitialized(self):
        """Returns the time since this command was initialized (in seconds).
        This function will work even if there is no specified timeout.
        
        :returns: the time since this command was initialized (in seconds).
        """
        with self.mutex:
            if self.startTime is None:
                return 0
            else:
                return Timer.getFPGATimestamp() - self.startTime

    def requires(self, subsystem):
        """This method specifies that the given Subsystem is used by this
        command.  This method is crucial to the functioning of the Command
        System in general.

        Note that the recommended way to call this method is in the
        constructor.

        :param subsystem: the :class:`.Subsystem` required
        """
        with self.mutex:
            if self.locked:
                raise ValueError("Can not add new requirement to command")
            if subsystem is None:
                raise ValueError("Subsystem must not be None.")
            self.requirements.add(subsystem)

    def removed(self):
        """Called when the command has been removed. This will call
        :meth:`interrupted` or :meth:`end`.
        """
        with self.mutex:
            if self.initialized:
                if self.isCanceled():
                    self.interrupted()
                    self._interrupted()
                else:
                    self.end()
                    self._end()
            self.initialized = False
            self.canceled = False
            self.running = False
        table = self.getTable()
        if table is not None:
            table.putBoolean("running", False)

    def run(self):
        """The run method is used internally to actually run the commands.
        
        :returns: whether or not the command should stay within the Scheduler.
        """
        with self.mutex:
            if not self.runWhenDisabled and self.parent is None and RobotState.isDisabled():
                self.cancel()
            if self.isCanceled():
                return False
            if not self.initialized:
                self.initialized = True
                self.startTiming()
                self._initialize()
                self.initialize()
            self._execute()
            self.execute()
            return not self.isFinished()

    def initialize(self):
        """The initialize method is called the first time this Command is run
        after being started.
        """
        pass

    def _initialize(self):
        """A shadow method called before initialize()."""
        pass

    def execute(self):
        """The execute method is called repeatedly until this Command either
        finishes or is canceled.
        """
        pass

    def _execute(self):
        """A shadow method called before execute()."""
        pass

    def isFinished(self):
        """Returns whether this command is finished.
        If it is, then the command will be removed and end() will be called.

        It may be useful for a team to reference the isTimedOut() method
        for time-sensitive commands, or override TimedCommand.

        If you do not specify isFinished in your command, the command will only
        end if interrupted or canceled. If you want a command that executes only
        once and then ends, override InstantCommand.

        :returns: whether this command is finished.
        :see: :meth:`isTimedOut`
        :see: :class: `.TimedCommand`
        :see: :class: `.InstantCommand`
        """
        return False

    def end(self):
        """Called when the command ended peacefully.  This is where you may
        want to wrap up loose ends, like shutting off a motor that was being
        used in the command.
        """
        pass

    def _end(self):
        """A shadow method called after end()."""
        pass

    def interrupted(self):
        """Called when the command ends because somebody called cancel() or
        another command shared the same requirements as this one, and booted
        it out.

        This is where you may want to wrap up loose ends, like shutting off a
        motor that was being used in the command.

        Generally, it is useful to simply call the end() method within this
        method, as done here.
        """
        self.end()

    def _interrupted(self):
        """A shadow method called after interrupted()."""
        pass

    def startTiming(self):
        """Called to indicate that the timer should start.
        This is called right before initialize() is, inside the run() method.
        """
        with self.mutex:
            self.startTime = Timer.getFPGATimestamp()

    def isTimedOut(self):
        """Returns whether or not the :meth:`timeSinceInitialized` method returns a
        number which is greater than or equal to the timeout for the command.
        If there is no timeout, this will always return false.
        
        :returns: whether the time has expired
        """
        with self.mutex:
            return (self.timeout != -1 and
                    self.timeSinceInitialized() >= self.timeout)

    def getRequirements(self):
        """Returns the requirements (as a set of Subsystems) of this command
        """
        with self.mutex:
            return self.requirements.copy()

    def lockChanges(self):
        """Prevents further changes from being made
        """
        with self.mutex:
            self.locked = True

    def setParent(self, parent):
        """Sets the parent of this command.  No actual change is made to the
        group.

        :param parent: the parent
        """
        with self.mutex:
            if self.parent is not None:
                raise ValueError("Can not give command to a command group after already being put in a command group")
            self.lockChanges()
            self.parent = parent
        table = self.getTable()
        if table is not None:
            table.putBoolean("isParented", True)
            
    def clearRequirements(self):
        """Clears list of subsystem requirements. This is only used by
        :class:`.ConditionalCommand` so cancelling the chosen command works properly
        in :class:`.CommandGroup`.
        """
        self.requirements = set()

    def start(self):
        """Starts up the command.  Gets the command ready to start.
        Note that the command will eventually start, however it will not
        necessarily do so immediately, and may in fact be canceled before
        initialize is even called.
        """
        with self.mutex:
            self.lockChanges()
            if self.parent is not None:
                raise ValueError("Can not start a command that is a part of a command group")
            Scheduler.getInstance().add(self)

    def startRunning(self):
        """This is used internally to mark that the command has been started.
        The lifecycle of a command is:

        * :meth:`startRunning` is called.
        * :meth:`run` is called (multiple times potentially)
        * :meth:`removed` is called

        It is very important that :meth:`startRunning` and :meth:`removed` be
        called in order or some assumptions of the code will be broken.
        """
        with self.mutex:
            self.running = True
            self.startTime = None
        table = self.getTable()
        if table is not None:
            table.putBoolean("running", True)

    def isRunning(self):
        """Returns whether or not the command is running.
        This may return true even if the command has just been canceled, as it
        may not have yet called :meth:`interrupted`.
        
        :returns: whether or not the command is running
        """
        with self.mutex:
            return self.running

    def cancel(self):
        """This will cancel the current command.

        This will cancel the current command eventually.  It can be called
        multiple times.  And it can be called when the command is not running.
        If the command is running though, then the command will be marked as
        canceled and eventually removed.

        .. warning:: A command can not be canceled if it is a part of a
                    :class:`.CommandGroup`, you must cancel the CommandGroup
                    instead.
        """
        if self.parent is not None:
            raise ValueError("Can not manually cancel a command in a command group")
        self._cancel()

    def _cancel(self):
        """This works like cancel(), except that it doesn't throw an exception
        if it is a part of a command group.  Should only be called by the
        parent command group.
        """
        with self.mutex:
            if self.isRunning():
                self.canceled = True

    def isCanceled(self):
        """Returns whether or not this has been canceled.
        
        :returns: whether or not this has been canceled
        """
        with self.mutex:
            return self.canceled

    def isInterruptible(self):
        """Returns whether or not this command can be interrupted.
        
        :returns: whether or not this command can be interrupted
        """
        with self.mutex:
            return self.interruptible

    def setInterruptible(self, interruptible):
        """Sets whether or not this command can be interrupted.
        
        :param interruptible: whether or not this command can be interrupted
        """
        with self.mutex:
            self.interruptible = interruptible

    def doesRequire(self, system):
        """Checks if the command requires the given :class:`.Subsystem`.
        
        :param system: the system
        :returns: whether or not the subsystem is required, or False if given
                  None.
        """
        with self.mutex:
            return system in self.requirements

    def getGroup(self):
        """Returns the :class:`.CommandGroup` that this command is a part of.
        Will return None if this Command is not in a group.
        
        :returns: the :class:`.CommandGroup` that this command is a part of
                  (or None if not in group)
        """
        with self.mutex:
            return self.parent

    def setRunWhenDisabled(self, run):
        """Sets whether or not this {@link Command} should run when the robot
        is disabled.

        By default a command will not run when the robot is disabled, and will
        in fact be canceled.

        :param run: whether or not this command should run when the robot is
                    disabled
        """
        with self.mutex:
            self.runWhenDisabled = run

    def willRunWhenDisabled(self):
        """Returns whether or not this Command will run when the robot is
        disabled, or if it will cancel itself.
        """
        with self.mutex:
            return self.runWhenDisabled

    def __str__(self):
        """The string representation for a Command is by default its name.
        
        :returns: the string representation of this object
        """
        return self.getName()

    def getSmartDashboardType(self):
        return "Command"

    def valueChanged(self, itable, key, value, isNew):
        if value:
            self.start()
        else:
            self.cancel()

    def initTable(self, table):
        oldtable = self.getTable()
        if oldtable is not None:
            oldtable.removeTableListener(self.valueChanged)
        super().initTable(table)
        if table is not None:
            table.putString("name", self.getName())
            table.putBoolean("running", self.isRunning())
            table.putBoolean("isParented", self.parent is not None)
            table.addTableListener(self.valueChanged, False, key="running")
